#!/usr/bin/env python

"""
RSVP instruction and serialisation classes.

Copyright (C) 2007, 2008, 2009 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.data import Attr, Const, Class, Function, Module
from micropython.program import Block, DataObject, DataValue

def name(attr):
    if isinstance(attr, Attr):
        return attr.name or "<unnamed>"
    elif isinstance(attr, (Block, Const)):
        return attr
    else:
        return attr.full_name() or "<unnamed>"

# Serialisation-related classes.

class RSVPObject:

    "A generic data object wrapper."

    def __init__(self, item):
        self.item = item

    def set_location(self, location, with_builtins):
        self.item.location = location
        return location + 1

    def finalise_location(self, with_builtins):
        pass

    def _finalise_location(self, with_builtins):

        """
        Set the code body location for items now that the code blocks have been
        positioned.
        """

        item = self.item

        if not with_builtins and item.module.name == "__builtins__" and item.astnode.doc is None:
            item.code_body_location = item.full_name()
        else:
            item.code_body_location = item.get_body_block().location

class RSVPAttr(RSVPObject):

    "A wrapper for attributes/values."

    def as_raw(self, objtable, paramtable, with_builtins):
        item = self.item
        return [
            DataValue(
                item.get_context() and item.get_context().location,
                item.get_value() and item.get_value().location
                )
            ]

class RSVPBlock(RSVPObject):

    "A wrapper for blocks."

    def set_location(self, location, with_builtins):
        item = self.item
        item.location = location
        return location + len(item.code)

    def as_raw(self, objtable, paramtable, with_builtins):
        return self.item.code

class RSVPClass(RSVPObject):

    "A wrapper for classes."

    def set_location(self, location, with_builtins):
        self.item.instance_template_location = location
        return RSVPObject.set_location(self, location + 1, with_builtins)

    def finalise_location(self, with_builtins):
        self._finalise_location(with_builtins)

    def as_raw(self, objtable, paramtable, with_builtins):
        item = self.item
        classcode = objtable.as_list().get_code(item.full_name())
        attrcode = objtable.get_index(item.full_name())

        # Include a template of an instance for use when instantiating classes.

        call_method = item.all_class_attributes().get("__call__")
        call_method_value = call_method and call_method.get_value()
        call_method_code_location = call_method_value and call_method_value.code_location
        call_method_funccode = call_method_value and paramtable.as_list().get_code(call_method_value.full_name())

        instantiator_funccode = paramtable.as_list().get_code(item.get_instantiator().full_name())

        # NOTE: The instantiator code is the first block of the class.

        with_instantiator = with_builtins or item.module.name != "__builtins__" or item.astnode.doc is not None

        if not with_instantiator:
            instantiator_code_location = item.full_name()
        else:
            instantiator_code_location = item.get_instantiator().blocks[0].location

        return [

            # Template instance...

            DataObject(
                classcode,
                attrcode, # is instance
                call_method_code_location,
                item.full_name(),
                len(item.instance_attributes()) + 1, # size
                call_method_funccode # funccode
                ),

            # Class...

            DataObject(
                classcode,
                None, # is not instance
                instantiator_code_location,
                item.full_name(),
                len(item.class_attributes()) + 1, # size
                instantiator_funccode # funccode
                )
            ]

class RSVPConst(RSVPObject):

    "A wrapper for constants."

    def set_location(self, location, with_builtins):
        location = RSVPObject.set_location(self, location, with_builtins)
        return location + len(self.raw_data())

    def as_raw(self, objtable, paramtable, with_builtins):
        item = self.item
        # NOTE: Need class details!
        return [
            DataObject(
                objtable.as_list().get_code(item.value_type_name()),
                objtable.get_index(item.value_type_name()), # is instance
                None,
                item.value_type_name(),
                1 # size
                )
            ] + self.raw_data()

    def raw_data(self):
        item = self.item
        value = item.get_value()
        # NOTE: Start simple and use single entries for most types.
        if item.value_type_name() in ("__builtins__.tuple", "__builtins__.list"):
            return [len(value)] + list(value)
        else:
            return [value]

class RSVPFunction(RSVPObject):

    "A wrapper for functions."

    def set_location(self, location, with_builtins):
        item = self.item
        location = RSVPObject.set_location(self, location, with_builtins)

        # Set the code location only where the code has been
        # generated.

        if not with_builtins and item.module.name == "__builtins__" and item.astnode.doc is None:
            item.code_location = item.full_name()

        # Skip any defaults for named functions.

        elif item.name is not None:
            item.code_location = location + len(item.defaults)

        # Skip any defaults for lambda functions.

        else:
            item.code_location = location

        return location

    def finalise_location(self, with_builtins):
        self._finalise_location(with_builtins)

    def as_raw(self, objtable, paramtable, with_builtins):
        item = self.item
        # NOTE: Need class and parameter details! Should arguably be an instance of types.FunctionType.
        return [
            DataObject(
                objtable.as_list().get_code("__builtins__.function"),
                objtable.get_index("__builtins__.function"), # is instance
                item.code_location,
                "__builtins__.function",
                len(item.defaults) + 1, # size (not accurate for lambda functions before instantiation)
                paramtable.as_list().get_code(item.full_name()) # funccode
                )
            ]

class RSVPModule(RSVPObject):

    "A wrapper for modules."

    def as_raw(self, objtable, paramtable, with_builtins):
        item = self.item
        return [
            DataObject(
                objtable.as_list().get_code(item.full_name()),
                None, # modules treated like classes
                None,
                item.full_name(),
                len(item.module_attributes()) + 1 # size
                )
            ]

# Serialisation-related data and functions.

def get_object(item):
    if isinstance(item, Attr):
        cls = RSVPAttr
    elif isinstance(item, Block):
        cls = RSVPBlock
    elif isinstance(item, Class):
        cls = RSVPClass
    elif isinstance(item, Const):
        cls = RSVPConst
    elif isinstance(item, Function):
        cls = RSVPFunction
    elif isinstance(item, Module):
        cls = RSVPModule
    else:
        cls = None

    if cls is not None:
        return cls(item)
    else:
        return None

# Instruction-related classes.

class Instruction:

    "A generic instruction."

    def __init__(self, attr=None):
        self.attr = attr
        self.input = None
        self.source = None # for storage instructions

    def copy(self):
        return self.__class__(self.attr)

    def __repr__(self):
        if self.attr is not None:
            return "%s(%r)%s" % (self.__class__.__name__, self.attr, self.show_input())
        else:
            return "%s()%s" % (self.__class__.__name__, self.show_input())

    def show_input(self):
        if self.input is not None:
            if self.source is not None:
                return " <- (%r, %r)" % (self.input, self.source)
            else:
                return " <- %r" % self.input
        elif self.source is not None:
            return " <-- %r" % self.source
        else:
            return ""

    def get_operand(self):
        return None

class FrameRelativeInstruction(Instruction):

    "An instruction operating on the current frame."

    def __repr__(self):
        return "%s(%r)%s" % (self.__class__.__name__, self.get_operand(), self.show_input())

    def get_operand(self):
        return self.attr.position

FR = FrameRelativeInstruction

class AddressRelativeInstruction(Instruction):

    "An instruction accessing an object's attribute."

    def __repr__(self):
        position = self.get_operand()
        if position is not None:
            return "%s(%r)%s # %s" % (self.__class__.__name__, position, self.show_input(), name(self.attr))
        else:
            return "%s(%r)%s" % (self.__class__.__name__, name(self.attr), self.show_input())

    def get_operand(self):
        return self.attr.position

AR = AddressRelativeInstruction

class AddressInstruction(Instruction):

    "An instruction loading an address directly."

    def __repr__(self):
        location, position, result = self.get_operands()
        if location is not None:
            return "%s(%r)%s # %r, %r (%s)" % (
                self.__class__.__name__, result, self.show_input(), location, position, name(self.attr))
        elif result is not None:
            return "%s(%r)%s # %s" % (
                self.__class__.__name__, result, self.show_input(), name(self.attr))
        else:
            return "%s(...)%s # %s" % (
                self.__class__.__name__, self.show_input(), name(self.attr))

    def get_operands(self):
        if isinstance(self.attr, Attr):
            position = self.attr.position
            location = self.attr.parent.location

            # NOTE: Unpositioned attributes are handled here.

            if location is not None and position is not None:
                result = location + position + 1
            else:
                location = self.attr.parent.name
                position = self.attr.name
                result = None
            return location, position, result
        else:
            return None, None, self.attr.location

    def get_operand(self):
        return self.get_operands()[-1]

Address = AddressInstruction

class TargetInstruction(Instruction):

    "An instruction loading the address of an invocation target."

    def __repr__(self):
        return "%s(%r) # %r" % (self.__class__.__name__, self.get_operand(), name(self.attr))

    def get_operand(self):
        return self.attr.code_body_location

Target = TargetInstruction

class ImmediateInstruction(Instruction):

    "An instruction employing a constant."

    def __repr__(self):
        return "%s(%r)%s" % (self.__class__.__name__, self.attr, self.show_input())

    def get_operand(self):
        return self.attr

Immediate = ImmediateInstruction

# Access to stored constant data.

class LoadConst(Address):           "Load the constant or module from the specified location."
class LoadClass(Address):           "Load the class from the specified location."
class LoadFunction(Address):        "Load the function from the specified location."

# Access within an invocation frame.

class LoadName(FR):                 "Load the current value from the given local attribute/variable."
class StoreName(FR):                "Store the source value into the given local attribute/variable."
class LoadTemp(Immediate):          "Load the current value from the given temporary location."
class StoreTemp(Immediate):         "Store the current value into the given temporary location."

# Access to static data.

class LoadAddress(Address):         "Load the current value from the given fixed attribute address."
class StoreAddress(Address):        "Store the source value into the given fixed attribute address."
class LoadAddressContext(Address):  "Load the current value from the given fixed attribute address, using the current value as context."
class StoreAddressContext(Address): "Store the current value into the given fixed attribute address, using the current value as context."
class LoadAddressContextCond(Address):
                                  """Load the current value from the given fixed attribute address, only using the current value as
                                     context if the attribute is compatible."""
class MakeInstance(Immediate):      "Make a new instance using the current value as a reference to a template."
class MakeFragment(Immediate):      "Make a new list fragment."

# Access to address-relative data. (LoadAttrIndexContext not defined.)

class LoadAttr(AR):                 "Load into the current value the given attribute of the object referenced by the current value."
class StoreAttr(AR):                "Store the source value into the given attribute of the object referenced by the current value."
class LoadAttrIndex(Immediate):     "Load into the current value the attribute of the current value with the given index."
class StoreAttrIndex(Immediate):    "Store the source value into the attribute of the current value with the given index."
class LoadAttrIndexContextCond(Immediate):
                                  """Load into the current value the attribute of the current value with the given index, only making the
                                     current value the context if the attribute is compatible."""

# Access to object details.

class LoadCallable(Instruction):    "Load the target of an invocation."
class StoreCallable(Instruction):   "Store the source value into the object referenced by the current value."

# Access to invocation frames in preparation.

class MakeFrame(Immediate):         "Make a new invocation frame."
class AdjustFrame(Immediate):       "Adjust the current invocation frame for corrected invocations."
class DropFrame(Instruction):       "Drop an invocation frame."
class StoreFrame(Immediate):        "Store the current value as an argument for the parameter with the given position."
class StoreFrameIndex(Immediate):   "Store the source value as an argument of the current value for the parameter with the given index."
class LoadContext(Instruction):     "Load the context of an invocation."

# Context-related tests.

class CheckContext(Instruction):    "Check to see if the context is valid."
class CheckClass(Instruction):      "Check the current value to determine whether it is a class."
class CheckInstance(Instruction):   "Check the current value as an instance of a specific class (used with 'self' in an invocation)."

# Access to frames upon invocation.

class CheckFrame(Immediate):        "Check the frame for the correct number of arguments."
class CheckExtra(Immediate):        "Ensure that the frame can provide extra arguments."
class FillDefaults(Immediate):      "Fill frame positions with defaults, if appropriate."
class ExtendFrame(Immediate):       "Extend the current frame for temporary storage use."
class CopyExtra(Immediate):         "Copy extra arguments into a separate sequence, starting from the given position."

# Invocation-related instructions, using a special result "register".

class JumpInFrame(Instruction):     "Jump, using the current locals, to the current callable."
class JumpWithFrame(Instruction):   "Jump, adopting the invocation frame, to the current callable."
class JumpWithFrameDirect(Target):  "Jump to the specified address, adopting the invocation frame."
class Return(Instruction):          "Return from a subprogram."
class LoadResult(Instruction):      "Load into the current value a returned value."
class StoreResult(Instruction):     "Store the current value as a value to be returned."

# Branch-related instructions.

class Jump(Address):                "Jump unconditionally."
class JumpIfFalse(Address):         "Jump if the last evaluation gave a false result."
class JumpIfTrue(Address):          "Jump if the last evaluation gave a true result."

# Exception-related instructions, using a special exception "register".

class LoadException(Instruction):   "Load the raised exception."
class StoreException(Instruction):  "Store the current object in the exception register."
class ClearException(Instruction):  "Reset the exception register."
class RaiseException(Instruction):  "Raise an exception, jumping to the active handler."
class PushHandler(Address):         "Push an exception handler onto the handler stack."
class PopHandler(Instruction):      "Pop an exception handler from the handler stack."
class CheckException(Instruction):  "Check the raised exception against another."

# Test instructions, operating on the boolean status register.

class TestIdentity(Instruction):    "Test whether the current value is identical to the source value, setting the boolean status."
class TestIdentityAddress(Address): "Test whether the current value is identical to the given address, setting the boolean status."
class InvertBoolean(Instruction):   "Invert the boolean status."

# Instructions which affect the current value. (LoadAttrIndexContext not defined.)

current_value_instructions = (
    LoadConst, LoadClass, LoadFunction, LoadName, LoadTemp,
    LoadAddress, LoadAddressContext, LoadAddressContextCond,
    LoadAttr, LoadAttrIndex, LoadAttrIndexContextCond,
    LoadCallable, LoadContext, LoadResult,
    LoadException, MakeInstance, MakeFragment,
    CopyExtra
    )

# Instructions which use the current value.

simple_input_user_instructions = (
    StoreTemp, StoreFrame, StoreResult, StoreException, # as the value being stored
    LoadAddressContext, LoadAddressContextCond,         # as the object being referenced
    LoadAttr, LoadAttrIndex, # LoadAttrIndexContext,    # as the object being referenced
    LoadAttrIndexContextCond,                           # as the object being referenced
    StoreAttr, StoreAttrIndex, StoreCallable,           # as the object being referenced
    StoreFrameIndex,                                    # as the object being referenced
    StoreAddressContext,                                # as the context
    LoadCallable,
    TestIdentity, TestIdentityAddress, CheckInstance,   # as one of the operands
    CheckException, CheckFrame, FillDefaults,
    MakeInstance,
    CheckContext, CheckClass,
    LoadContext                                         # as the object providing the result
    )

# vim: tabstop=4 expandtab shiftwidth=4
